% Calculate orientation the beach profile at each unique site and save to .mat file. Orientation is
% the number of degrees, anticlockwise from east, perpendicular to the shoreline (pointing towards
% land).

%% Setup
% Needs the following coastal tools:
addpath('J:\Coastal\Tools\MALT Logspiral Transformation')
addpath('J:\Coastal\Tools\Coordinate Transformations')

clear 
clc

%% Options
% Where is the profiles file located? This should contain a structure including the .lat and .lon
% for each analysed cross section
profilesFile = '..\..\data\raw\processed_shorelines\profiles.mat';

% Where should we store the processed beach orientations?
outputFile = '..\..\data\raw\processed_shorelines\orientations.mat';

% How far in meters does the profile extend towards land and sea? Used to provide end points of the
% cross section
distance = 200;


%% Script

% Load profile data, this is where we want to calculate orientations.
warning('off','all')
data = load(profilesFile);
data = data.data;

% Save results to variable
output = [];

for ii = 1:length(data)
  disp(num2str(ii))
  lat = data(ii).lat;
  lon = data(ii).lon;
  beach = data(ii).site{1};
  [x,y,utmzone] = deg2utm(lat,lon);

  if strcmp(beach, 'BOOM') == 1 || strcmp(beach, 'HARGn') == 1 || strcmp(beach, 'BILG') == 1 || strcmp(beach, 'HARGs') == 1 || strcmp(beach, 'DEEWHYn') == 1
    % These are straight beaches, load the transformation file directly and read the rotation angle.
    parameterDir = 'J:\Coastal\Tools\MALT Logspiral Transformation';
    parameterFile = [parameterDir, filesep, beach, '.mat'];
    parameterMat = load(parameterFile);
    fields = fieldnames(parameterMat);
    field = fields(1,1);
    site = getfield(parameterMat, field{1});
    rot_angle = site.rot_angle;  % Angle of the shoreline counter clockwise from east
    
    % Figure out end points in utm coordinates

    x_land = x - distance * cos(deg2rad(rot_angle));
    y_land = y + distance * sin(deg2rad(rot_angle));
    x_sea = x + distance * cos(deg2rad(rot_angle));
    y_sea = y - distance * sin(deg2rad(rot_angle));
    
    [lat_land,lon_land] = utm2deg(x_land,y_land,utmzone);
    [lat_sea,lon_sea] = utm2deg(x_land,y_land,utmzone);
    
    row.lat_center = lat;
    row.lon_center = lon;
    row.lat_land = lat_land;
    row.lon_land = lat_land;
    row.lat_sea = lat_sea;
    row.lon_sea = lon_sea;
    row.orientation = rot_angle + 90;  % Tangent to shoreline towards line 
    row.beach = beach;
    output = [output; row];
    continue
  end
    
%   if strcmp(beach, 'AVOCAs') == 1
%     % negative solution. Talk to Mitch
%     continue
%   end
  
  % Get the sp log spiral transformed coordinates
  xyz.x = x;
  xyz.y = y;
  xyz.z = 0;
  out = xyz2spz(xyz,beach);

  % Define new bounds towards land and towards sea
  % Distance is the number of m in the section to go towards
  distance = 200;
  s = out.s;
  p_center = out.p;

  % Convert landwards and seawrds point back to lat/lon
  spz_land.s = out.s;
  spz_land.p = out.p-distance;
  spz_land.z = 0;
  xyz_land = spz2xyz(spz_land,beach);
  [lat_land,lon_land] = utm2deg(xyz_land.x,xyz_land.y,utmzone);

  spz_sea.s = out.s;
  spz_sea.p = out.p+distance;
  spz_sea.z = 0;
  xyz_sea = spz2xyz(spz_sea,beach);
  [lat_sea,lon_sea] = utm2deg(xyz_sea.x,xyz_sea.y,utmzone);

  % Orientation in degrees anticlockwise from east, pointing towards land
  try
    orientation = radtodeg(atan2((xyz_land.y - xyz_sea.y), (xyz_land.x - xyz_sea.x)));
  catch
    disp(['Cannot calculate orientation: ' beach])
    continue
  end
  
  row.lat_center = lat;
  row.lon_center = lon;
  row.lat_land = lat_land;
  row.lon_land = lat_land;
  row.lat_sea = lat_sea;
  row.lon_sea = lon_sea;
  row.orientation = orientation;
  row.beach = beach;
  output = [output; row];
end
warning('on','all')

save(outputFile','output','-v7')
